import ssl
from unittest import mock

import pytest  # noqa
import trustme  # noqa
from python_socks.async_.asyncio._resolver import Resolver as AsyncioResolver  # noqa

from tests.config import (
    PROXY_HOST_IPV4,
    PROXY_HOST_IPV6,
    SOCKS5_PROXY_PORT,
    LOGIN,
    PASSWORD,
    SKIP_IPV6_TESTS,
    HTTP_PROXY_PORT,
    SOCKS4_PORT_NO_AUTH,
    SOCKS4_PROXY_PORT,
    SOCKS5_PROXY_PORT_NO_AUTH,
    TEST_PORT_IPV4,
    TEST_PORT_IPV6,
    TEST_HOST_IPV4,
    TEST_HOST_IPV6,
    TEST_PORT_IPV4_HTTPS,
    TEST_HOST_NAME_IPV4,
    TEST_HOST_NAME_IPV6,
)
from tests.http_server import HttpServer, HttpServerConfig
from tests.mocks import async_resolve_factory
from tests.proxy_server import ProxyConfig, ProxyServer
from tests.utils import wait_until_connectable


@pytest.fixture(scope='session')
def target_ssl_ca() -> trustme.CA:
    return trustme.CA()


@pytest.fixture(scope='session')
def target_ssl_cert(target_ssl_ca) -> trustme.LeafCert:
    return target_ssl_ca.issue_cert(
        'localhost',
        TEST_HOST_IPV4,
        TEST_HOST_IPV6,
        TEST_HOST_NAME_IPV4,
        TEST_HOST_NAME_IPV6,
    )


@pytest.fixture(scope='session')
def target_ssl_certfile(target_ssl_cert):
    with target_ssl_cert.cert_chain_pems[0].tempfile() as cert_path:
        yield cert_path


@pytest.fixture(scope='session')
def target_ssl_keyfile(target_ssl_cert):
    with target_ssl_cert.private_key_pem.tempfile() as private_key_path:
        yield private_key_path


@pytest.fixture(scope='session')
def target_ssl_context(target_ssl_ca) -> ssl.SSLContext:
    ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    ssl_ctx.verify_mode = ssl.CERT_REQUIRED
    ssl_ctx.check_hostname = True
    target_ssl_ca.configure_trust(ssl_ctx)
    return ssl_ctx


@pytest.fixture(scope='session', autouse=True)
def patch_resolvers():
    with mock.patch.object(
        AsyncioResolver, attribute='resolve', new=async_resolve_factory(AsyncioResolver)
    ):
        yield None


@pytest.fixture(scope='session', autouse=True)
def proxy_server():
    config = [
        ProxyConfig(
            proxy_type='http',
            host=PROXY_HOST_IPV4,
            port=HTTP_PROXY_PORT,
            username=LOGIN,
            password=PASSWORD,
        ),
        ProxyConfig(
            proxy_type='socks4',
            host=PROXY_HOST_IPV4,
            port=SOCKS4_PROXY_PORT,
            username=LOGIN,
            password=None,
        ),
        ProxyConfig(
            proxy_type='socks4',
            host=PROXY_HOST_IPV4,
            port=SOCKS4_PORT_NO_AUTH,
            username=None,
            password=None,
        ),
        ProxyConfig(
            proxy_type='socks5',
            host=PROXY_HOST_IPV4,
            port=SOCKS5_PROXY_PORT,
            username=LOGIN,
            password=PASSWORD,
        ),
        ProxyConfig(
            proxy_type='socks5',
            host=PROXY_HOST_IPV4,
            port=SOCKS5_PROXY_PORT_NO_AUTH,
            username=None,
            password=None,
        ),
    ]

    if not SKIP_IPV6_TESTS:
        config.append(
            ProxyConfig(
                proxy_type='socks5',
                host=PROXY_HOST_IPV6,
                port=SOCKS5_PROXY_PORT,
                username=LOGIN,
                password=PASSWORD,
            ),
        )

    server = ProxyServer(config=config)
    server.start()
    for cfg in config:
        wait_until_connectable(host=cfg.host, port=cfg.port, timeout=10)

    yield None

    server.terminate()


@pytest.fixture(scope='session', autouse=True)
def web_server(target_ssl_certfile, target_ssl_keyfile):
    config = [
        HttpServerConfig(host=TEST_HOST_IPV4, port=TEST_PORT_IPV4),
        HttpServerConfig(
            host=TEST_HOST_IPV4,
            port=TEST_PORT_IPV4_HTTPS,
            certfile=target_ssl_certfile,
            keyfile=target_ssl_keyfile,
        ),
    ]

    if not SKIP_IPV6_TESTS:
        config.append(HttpServerConfig(host=TEST_HOST_IPV6, port=TEST_PORT_IPV6))

    server = HttpServer(config=config)
    server.start()
    for cfg in config:
        server.wait_until_connectable(host=cfg.host, port=cfg.port)

    yield None

    server.terminate()



import asyncio
import ssl

import aiohttp
import pytest  # noqa
from aiohttp import ClientResponse, TCPConnector
from yarl import URL  # noqa

from aiohttp_socks import (
    ProxyType,
    ProxyConnector,
    ChainProxyConnector,
    ProxyInfo,
    ProxyError,
    ProxyConnectionError,
    ProxyTimeoutError,
    open_connection,
    create_connection,
)
from tests.config import (
    TEST_URL_IPV4,
    SOCKS5_IPV4_URL,
    PROXY_HOST_IPV4,
    SOCKS5_PROXY_PORT,
    LOGIN,
    PASSWORD,
    TEST_URL_IPV4_DELAY,
    SKIP_IPV6_TESTS,
    SOCKS5_IPV6_URL,
    SOCKS4_URL,
    HTTP_PROXY_URL,
    SOCKS4_PROXY_PORT,
    HTTP_PROXY_PORT,
    TEST_URL_IPV4_HTTPS,
)


async def fetch(
    connector: TCPConnector,
    url: str,
    timeout=None,
    ssl_context=None,
) -> ClientResponse:
    url = URL(url)

    if url.scheme == 'https':
        dest_ssl = ssl_context
    else:
        dest_ssl = None

    async with aiohttp.ClientSession(connector=connector) as session:
        async with session.get(url, ssl=dest_ssl, timeout=timeout) as resp:
            return resp


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks5_proxy_ipv4(url, rdns, target_ssl_context):
    connector = ProxyConnector.from_url(SOCKS5_IPV4_URL, rdns=rdns)
    res = await fetch(
        connector=connector,
        url=url,
        ssl_context=target_ssl_context,
    )
    assert res.status == 200


@pytest.mark.asyncio
async def test_socks5_proxy_with_invalid_credentials(target_ssl_context):
    connector = ProxyConnector(
        proxy_type=ProxyType.SOCKS5,
        host=PROXY_HOST_IPV4,
        port=SOCKS5_PROXY_PORT,
        username=LOGIN,
        password=PASSWORD + 'aaa',
    )
    with pytest.raises(ProxyError):
        await fetch(
            connector=connector,
            url=TEST_URL_IPV4,
            ssl_context=target_ssl_context,
        )


@pytest.mark.asyncio
async def test_socks5_proxy_with_timeout(target_ssl_context):
    connector = ProxyConnector(
        proxy_type=ProxyType.SOCKS5,
        host=PROXY_HOST_IPV4,
        port=SOCKS5_PROXY_PORT,
        username=LOGIN,
        password=PASSWORD,
    )
    with pytest.raises(asyncio.TimeoutError):
        await fetch(
            connector=connector,
            url=TEST_URL_IPV4_DELAY,
            timeout=1,
            ssl_context=target_ssl_context,
        )


@pytest.mark.asyncio
async def test_socks5_proxy_with_proxy_connect_timeout(target_ssl_context):
    connector = ProxyConnector.from_url(SOCKS5_IPV4_URL)
    timeout = aiohttp.ClientTimeout(total=32, sock_connect=0.001)
    with pytest.raises(ProxyTimeoutError):
        await fetch(
            connector=connector,
            url=TEST_URL_IPV4,
            timeout=timeout,
            ssl_context=target_ssl_context,
        )


@pytest.mark.asyncio
async def test_socks5_proxy_with_invalid_proxy_port(unused_tcp_port, target_ssl_context):
    connector = ProxyConnector(
        proxy_type=ProxyType.SOCKS5,
        host=PROXY_HOST_IPV4,
        port=unused_tcp_port,
        username=LOGIN,
        password=PASSWORD,
    )
    with pytest.raises(ProxyConnectionError):
        await fetch(
            connector=connector,
            url=TEST_URL_IPV4,
            ssl_context=target_ssl_context,
        )


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.skipif(SKIP_IPV6_TESTS, reason="TravisCI doesn't support ipv6")
@pytest.mark.asyncio
async def test_socks5_proxy_ipv6(url, target_ssl_context):
    connector = ProxyConnector.from_url(SOCKS5_IPV6_URL)
    res = await fetch(
        connector=connector,
        url=url,
        ssl_context=target_ssl_context,
    )
    assert res.status == 200


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks4_proxy(url, rdns, target_ssl_context):
    connector = ProxyConnector.from_url(
        SOCKS4_URL,
        rdns=rdns,
    )
    res = await fetch(
        connector=connector,
        url=url,
        ssl_context=target_ssl_context,
    )
    assert res.status == 200


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.asyncio
async def test_http_proxy(url, target_ssl_context):
    connector = ProxyConnector.from_url(HTTP_PROXY_URL)
    res = await fetch(
        connector=connector,
        url=url,
        ssl_context=target_ssl_context,
    )
    assert res.status == 200


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.asyncio
async def test_chain_proxy_from_url(url, target_ssl_context):
    connector = ChainProxyConnector.from_urls(
        [SOCKS5_IPV4_URL, SOCKS4_URL, HTTP_PROXY_URL]
    )
    res = await fetch(
        connector=connector,
        url=url,
        ssl_context=target_ssl_context,
    )
    assert res.status == 200


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_chain_proxy_ctor(url, rdns, target_ssl_context):
    connector = ChainProxyConnector(
        [
            ProxyInfo(
                proxy_type=ProxyType.SOCKS5,
                host=PROXY_HOST_IPV4,
                port=SOCKS5_PROXY_PORT,
                username=LOGIN,
                password=PASSWORD,
                rdns=rdns,
            ),
            ProxyInfo(
                proxy_type=ProxyType.SOCKS4,
                host=PROXY_HOST_IPV4,
                port=SOCKS4_PROXY_PORT,
                username=LOGIN,
                rdns=rdns,
            ),
            ProxyInfo(
                proxy_type=ProxyType.HTTP,
                host=PROXY_HOST_IPV4,
                port=HTTP_PROXY_PORT,
                username=LOGIN,
                password=PASSWORD,
            ),
        ]
    )
    res = await fetch(
        connector=connector,
        url=url,
        ssl_context=target_ssl_context,
    )
    assert res.status == 200


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks5_open_connection(url, rdns, target_ssl_context):
    url = URL(url)

    ssl_context = None
    if url.scheme == 'https':
        ssl_context = target_ssl_context

    reader, writer = await open_connection(
        proxy_url=SOCKS5_IPV4_URL,
        host=url.host,
        port=url.port,
        ssl=ssl_context,
        server_hostname=url.host if ssl_context else None,
        rdns=rdns,
    )
    request = "GET %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n" % (
        url.path_qs,
        url.host,
    )

    writer.write(request.encode())
    response = await reader.read(-1)
    assert b'200 OK' in response


@pytest.mark.parametrize('url', (TEST_URL_IPV4, TEST_URL_IPV4_HTTPS))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks5_http_create_connection(
    url: str,
    rdns: bool,
    target_ssl_context: ssl.SSLContext,
):
    url = URL(url)

    ssl_context = None
    if url.scheme == 'https':
        ssl_context = target_ssl_context

    event_loop = asyncio.get_running_loop()
    reader = asyncio.StreamReader(loop=event_loop)
    protocol = asyncio.StreamReaderProtocol(reader, loop=event_loop)

    transport, _ = await create_connection(
        proxy_url=SOCKS5_IPV4_URL,
        protocol_factory=lambda: protocol,
        host=url.host,
        port=url.port,
        ssl=ssl_context,
        server_hostname=url.host if ssl_context else None,
        rdns=rdns,
    )

    writer = asyncio.StreamWriter(transport, protocol, reader, event_loop)

    request = "GET %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n" % (
        url.path_qs,
        url.host,
    )

    writer.write(request.encode())
    response = await reader.read(-1)
    assert b'200 OK' in response



import ssl
import time

import flask  # noqa
from flask import request  # noqa

app = flask.Flask(__name__)


@app.route('/ip')
def ip():
    return request.remote_addr


@app.route('/delay/<int:seconds>')
def delay(seconds):
    time.sleep(seconds)
    return 'ok'


def run_app(host: str, port: int, certfile: str = None, keyfile: str = None):
    if certfile and keyfile:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        ssl_context.load_cert_chain(certfile, keyfile)
    else:
        ssl_context = None

    print('Starting http server on {}:{}...'.format(host, port))
    app.run(debug=False, host=host, port=port, threaded=True,
            ssl_context=ssl_context)



import socket

from tests.config import (
    TEST_HOST_NAME_IPV4,
    PROXY_HOST_NAME_IPV4,
    TEST_HOST_NAME_IPV6,
    PROXY_HOST_NAME_IPV6,
)


def getaddrinfo_sync_mock():
    _orig_getaddrinfo = socket.getaddrinfo

    def getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
        if host in (TEST_HOST_NAME_IPV4, PROXY_HOST_NAME_IPV4):
            return [(socket.AF_INET, socket.SOCK_STREAM, 6, '', ('127.0.0.1', port))]

        if host in (TEST_HOST_NAME_IPV6, PROXY_HOST_NAME_IPV6):
            return [(socket.AF_INET6, socket.SOCK_STREAM, 6, '', ('::1', port, 0, 0))]

        return _orig_getaddrinfo(host, port, family, type, proto, flags)

    return getaddrinfo


def getaddrinfo_async_mock(origin_getaddrinfo):
    async def getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
        if host in (TEST_HOST_NAME_IPV4, PROXY_HOST_NAME_IPV4):
            return [(socket.AF_INET, socket.SOCK_STREAM, 6, '', ('127.0.0.1', port))]

        if host in (TEST_HOST_NAME_IPV6, PROXY_HOST_NAME_IPV6):
            return [(socket.AF_INET6, socket.SOCK_STREAM, 6, '', ('::1', port, 0, 0))]

        return await origin_getaddrinfo(
            host,
            port,
            family=family,
            type=type,
            proto=proto,
            flags=flags,
        )

    return getaddrinfo


def _resolve_local(host):
    if host in (TEST_HOST_NAME_IPV4, PROXY_HOST_NAME_IPV4):
        return socket.AF_INET, '127.0.0.1'

    if host in (TEST_HOST_NAME_IPV6, PROXY_HOST_NAME_IPV6):
        return socket.AF_INET6, '::1'

    return None


def sync_resolve_factory(cls):
    original_resolver = cls.resolve

    def new_resolver(self, host, port=0, family=socket.AF_UNSPEC):
        res = _resolve_local(host)

        if res is not None:
            return res

        return original_resolver(self, host=host, port=port, family=family)

    return new_resolver


def async_resolve_factory(cls):
    original_resolver = cls.resolve

    async def new_resolver(self, host, port=0, family=socket.AF_UNSPEC):
        res = _resolve_local(host)

        if res is not None:
            return res

        return await original_resolver(self, host=host, port=port, family=family)

    return new_resolver
